/*
 *      syd  -- simulate a loblolly 3 generation pedigree
 *
 *	usage: syd
 *
 *	Uses configuration file syd.conf.
 *	If syd.conf does not exist, create syd.conf and exit.
 *
 *	Bradley K. Sherman
 *	bks@s27w007.pswfs.gov || bks@netcom.com
 *	Dendrome Project, Institute of Forest Genetics
 *	PO Box 245, Berkeley, CA 94701
 *
 *      --------Notes-------------
 *		
 *	n.b. Locus A_99999 is at position .99999 on Chromosome A
 *
 */

/* include's might need to be complexified for porting */
#include <stdio.h>
#include <assert.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/file.h>

/* Loci per chromosome */
#define MAX_L_PER_C 500
#define MAX_PROGENY 500
#define MAX_CHROMOSOME 500
#define MAX_CROSSOVER 10
#define MAX_LOCUS_POS .99999



/* A TREE has Chromosomes, and loci.  */
typedef struct tree {
        char *hapmom[MAX_CHROMOSOME];
        char *hapdad[MAX_CHROMOSOME];
} TREE;

static char SCCS[] = "@(#)syd.c	1.11 1/4/96 syd Dendrome project  --bks";

char *Progname;
TREE *Mom_mom;
TREE *Mom_dad;
TREE *Dad_mom;
TREE *Dad_dad;
TREE *Mom;
TREE *Dad;
TREE *Progeny[MAX_PROGENY];
double Position[MAX_CHROMOSOME][MAX_L_PER_C];

double R_delta; /* To map PRNG integers to 0.0 <= x < 1.0 */

/* On Each chromosome we track loci at these positions */
char Datafile[100] = "syd.dat";
char Logfile[100] = "syd.log";
char Configfile[] = "syd.conf";
FILE *Logfp;

/* User defined parameters/options */
char *Allele;
char *M_chiasma;
char *P_chiasma;
int L_per_c = 0;  /* Loci per Chromosome */
int Log = 0;
int N_progeny = 0;
int N_chromosome = 0;
int Phase_known = 0;
int Randloci = 0;
int Seed = 0;

double rand_pos();
char *crossover();


main( argc, argv )
int argc;
char *argv[];
{
	time_t tloc; 

	Progname = argv[0];
        init( argc, argv );
        go();

	printf( "\n%s complete.\nConfig file is %s.\nData file is %s.\n",
		Progname, Configfile, Datafile );
	if ( Log ) {
		printf( "Detailed history is in %s.\n", Logfile );
		time( &tloc );
		fprintf( Logfp, "%s end %s", Progname, ctime( &tloc ) );
	}
}

int
dcomp( d1, d2 )
double *d1;
double *d2;
{
        int i;
        return ( ( *d1 > *d2) ? 1 : -1 );
}

go()
{
        int i;
	if ( Randloci )
        	rand_select_loci(); /* XXX may be disabled */
	else
        	select_loci();
        if ( Log ) fprintf(Logfp, ":Generate Mom's Mom:\n");
        generate( Mom_mom );
        if ( Log ) fprintf(Logfp, ":Generate Mom's Dad:\n");
        generate( Mom_dad );
        if ( Log ) fprintf(Logfp, ":Generate Dad's Mom:\n");
        generate( Dad_mom );
        if ( Log ) fprintf(Logfp, ":Generate Dad's Dad:\n");
        generate( Dad_dad );
        if ( Log ) fprintf(Logfp, ":Mate for Mom:\n");
        mate( Mom, Mom_mom, Mom_dad );
        mate( Dad, Dad_mom, Dad_dad );
        for ( i = 0; i < N_progeny; i++ ) {
		if ( Log > 10 )
			fprintf ( Logfp, ":Mate for progeny %d:\n", i );
                mate( Progeny[i], Mom, Dad );
	}

        write_data();
}

write_data()
{
        FILE *fp;
        int i, c, j;
        char lname[BUFSIZ];

        if ( NULL == ( fp = fopen( Datafile, "w+") ) )
                perror( Datafile );
	/* Write header line --added 11/22/95 */
	fprintf( fp, "Locus" );
	fprintf( fp, " moms-mom");
	fprintf( fp, " moms-dad");
	fprintf( fp, " mom");
	fprintf( fp, " dad");
	fprintf( fp, " dads-mom");
	fprintf( fp, " dads-dad");
        for( j = 0; j < N_progeny; j++ ) {
		fprintf( fp, " p-%d", j + 1 );
	}
	fprintf( fp, "\n");
        for ( c = 0; c < N_chromosome; c++ ) {
                for ( i = 0; i < L_per_c; i++ ) {
                        sprintf( lname, "%0.5f", Position[c][i] );
                        fprintf( fp, "%c_%s ", 'A' + c, lname + 2 );

                        if ( Phase_known ) {
                                fprintf( fp, "%c%c ", 
                                        Mom_mom->hapmom[c][i],
                                        Mom_mom->hapdad[c][i] );
                                fprintf( fp, "%c%c ", 
                                        Mom_dad->hapmom[c][i],
                                        Mom_dad->hapdad[c][i] );
                                fprintf( fp, "%c%c ", 
                                        Mom->hapmom[c][i],
                                        Mom->hapdad[c][i] );
                                fprintf( fp, "%c%c ", 
                                        Dad->hapmom[c][i],
                                        Dad->hapdad[c][i] );
                                fprintf( fp, "%c%c ", 
                                        Dad_mom->hapmom[c][i],
                                        Dad_mom->hapdad[c][i] );
                                fprintf( fp, "%c%c ", 
                                        Dad_dad->hapmom[c][i],
                                        Dad_dad->hapdad[c][i] );
                                for( j = 0; j < N_progeny; j++ ) {
                                        fprintf( fp, "%c%c", 
                                                Progeny[j]->hapmom[c][i],
                                                Progeny[j]->hapdad[c][i] );
					if ( j != N_progeny - 1 )
						fprintf( fp, " " );
                                }
                        } else {
                                rand_write( fp, c, i );
                        }

                        fprintf( fp, "\n" );
                }
        }
        fclose( fp );
}

/*
 *	Mix up the alleles so phase is not obvious (consistent)
 */
rand_write( fp, c, i )
FILE *fp;
int c;
int i;
{
        int j;

	/* Leave grandparents alone, randomize other allele pairs */
        fprintf(fp,"%c%c ",
        	Mom_mom->hapmom[c][i],
                Mom_mom->hapdad[c][i]);
        fprintf( fp, "%c%c ", 
                Mom_dad->hapmom[c][i],
                Mom_dad->hapdad[c][i] );

        if ( random() % 2 )
                fprintf( fp, "%c%c ", 
                        Mom->hapdad[c][i],
                        Mom->hapmom[c][i] );
        else
                fprintf( fp, "%c%c ", 
                        Mom->hapmom[c][i],
                        Mom->hapdad[c][i] );

        if ( random() % 2 )
                fprintf( fp, "%c%c ", 
                        Dad->hapmom[c][i],
                        Dad->hapdad[c][i] );
        else
                fprintf( fp, "%c%c ", 
                        Dad->hapdad[c][i],
                        Dad->hapmom[c][i] );

       fprintf( fp, "%c%c ", 
                Dad_mom->hapmom[c][i],
                Dad_mom->hapdad[c][i] );

       fprintf( fp, "%c%c ", 
                Dad_dad->hapmom[c][i],
                Dad_dad->hapdad[c][i] );


        for( j = 0; j < N_progeny; j++ ) {
                if ( random() % 2 ) {
                        fprintf( fp, "%c%c", 
                                Progeny[j]->hapdad[c][i],
                                Progeny[j]->hapmom[c][i] );
		} else {
                        fprintf( fp, "%c%c", 
                                Progeny[j]->hapmom[c][i],
                                Progeny[j]->hapdad[c][i] );
		}
		if ( j != N_progeny - 1 )
			fprintf( fp, " " );
        }

}

/*
 *	Take mom and dad's chromosomes to produce kid's
 */
mate( kid, mom, dad )
TREE *kid;
TREE *mom;
TREE *dad;
{
        if ( Log > 10 ) fprintf(Logfp, ":Haploid contribution from Mother:\n" );
        hapmate( kid, mom, 0 );
        if ( Log > 10 ) fprintf(Logfp, ":Haploid contribution from Father:\n" );
        hapmate( kid, dad, 1 );
        show_tree( kid );
}

/*
 *	Get the contribution from one parent
 */
hapmate( kid, par, whichparent )
TREE *kid;
TREE *par;
int whichparent; /* 0 = mom, 1 = dad */
{
        int c;
        int cname;
        int nc;
        char hap0[MAX_L_PER_C];
        char hap1[MAX_L_PER_C];
        char *s;
        for ( c = 0; c < N_chromosome; c++ ) { /* For each chromosome */
                cname = 'A' + c;
		if ( whichparent == 0 )
                	nc = ncross( P_chiasma ); /* how many crossovers */
		else
			nc = ncross( P_chiasma );
                if ( Log > 10 )
			fprintf( Logfp, ":%c %d crossovers:\n", cname, nc );
                if ( random() % 2 ) /* Random start c.f. 0 crossovers */
                        s = crossover( c, nc, par->hapmom[c], par->hapdad[c] );
                else
                        s = crossover( c, nc, par->hapdad[c], par->hapmom[c] );
                if ( whichparent == 0 )
                        strcpy( kid->hapmom[c], s );
                else
                        strcpy( kid->hapdad[c], s );
        }
}

/*
 *      c is chromosome number
 *      nc is number of chiasmata; hap0, hap1 are chromosomes
 */
char *
crossover( chrom, nc, hap0, hap1 )
int chrom;
int nc;
char *hap0;
char *hap1;
{
        int i, copy, locus, chi;
        double pos;
        static char daughter[MAX_L_PER_C];
        double chiasma[MAX_CROSSOVER + 1];


        strcpy( daughter, hap0 );

        if ( nc > 0 ) {
                for ( i = 0; i < nc; i++ ) {
                        chiasma[i] = rand_pos();
                }
                qsort( chiasma, nc, sizeof(double), dcomp);
                /* XXX should see if chiasma are too close */
                if ( Log ) {
			fprintf( Logfp, "Chiasmata" );
                	for ( i = 0; i < nc; i++ ) {
                        	fprintf( Logfp, " %5d",
			 		(int)(100000.0 * chiasma[i]) );
                	}
                	fprintf( Logfp, "\n" );
		}
                chiasma[nc] = 1.0; /* sentinel XXX */

                copy = 0;
                chi = 0;
                for( locus = 0; locus < L_per_c; locus++ ) {
                        pos = Position[chrom][locus];
                        if ( copy == 0 ) {
                                if ( pos >= chiasma[chi] ) {
                                        daughter[locus] = hap1[locus];
                                        copy = 1;
                                        ++chi;
                                }
                        } else {
                                if ( pos >= chiasma[chi] ) {
                                        copy = 0;
                                        ++chi;
                                } else {
                                        daughter[locus] = hap1[locus];
                                }
                                        
                        }
                }
        }
        /* printf( "%d crossovers\n", nc ); */
        /* printf( "%s\n%s\n%s\n", hap0, hap1, daughter ); */

        return daughter;
}

/*
 *	Create a tree de novo (grandparents)
 */
generate( t )
TREE *t;
{
        int i, c;
	char *m, *d;
        for ( c = 0; c < N_chromosome; c++ ) {
		m = t->hapmom[c];
		d = t->hapdad[c];
                for ( i = 0; i < L_per_c; i++ )  {
                        *(m + i) = rand_allele();
                        *(d + i) = rand_allele();
                }
                *( m + L_per_c ) = '\0';
                *( d + L_per_c ) = '\0';
        }

        if ( Log > 10 )
                show_tree( t );
}

show_tree( t )
TREE *t;
{
        int i, c;
        for ( c = 0; c < N_chromosome; c++ ) {
                if (Log) fprintf( Logfp, "%c %s\n", 'A' + c, t->hapmom[c] );
                if (Log) fprintf( Logfp, "%c %s\n", 'A' + c, t->hapdad[c] );
        }
}

/* return an allele from the set of alleles possible */
rand_allele()
{

        int i, j;
        i = strlen( Allele );
        j = random() % i;
        return Allele[j];
}

/* Select loci at fixed points, always equidistant */
select_loci()
{
        int c, i;
	double delta;
	double telo0, telo1;
	double pos;

	assert( L_per_c > 1 );
	delta = 1.0 / (L_per_c - 1);

        for ( i = 0, pos = 0.0; i < L_per_c - 1; i++, pos += delta ) {
        	for ( c = 0; c < N_chromosome; c++ ) {
                        Position[c][i] = pos;
                }
        }
	for ( c = 0; c < N_chromosome; c++ ){
		Position[c][L_per_c - 1] = MAX_LOCUS_POS;
	}
        log_map();
}
/* Select loci at random places. May be disabled */
rand_select_loci()
{
        int c, i;
        double d;
        for ( c = 0; c < N_chromosome; c++ ) {
                for ( i = 0; i < L_per_c; i++ ) {
                        d = rand_pos();
                        assert( (d < 1.0) && (d >= 0.0) );
                        /* printf( "%d_%0.6f\n", c, Position[c][i] ); */
                        Position[c][i] = d;
                }
        }
        sort_loci();
        log_map();
}

/* Write the loci to the log */
log_map()
{
        int c, i;
        char cname[BUFSIZ];
        char lname[BUFSIZ];
        if ( Log  == 0 ) return;
	fprintf( Logfp, ":The Map:\n" );
        for ( c = 0; c < N_chromosome; c++ ) {
                cname[0] = 'A' + c;
                cname[1] = '\0';
               	for ( i = 0; i < L_per_c; i++ ) {
			sprintf( lname, "%0.5f", Position[c][i] );
                       	fprintf( Logfp, "%s_%s\n", cname, lname + 2 );
               	}
        }
}


sort_loci()
{
        int c, i;
        for ( c = 0; c < N_chromosome; c++ ) {
                qsort( &(Position[c][0]), L_per_c, sizeof(double), dcomp);
#ifdef NEVER
                for ( i = 0; i < L_per_c; i++ )
                        printf( "%.20f\n", Position[c][i] );
                printf( "-------\n" );
#endif
        }
}


/* Return a double 0.0 <= d < 1.0 */
double
rand_pos()
{
        double p;
        do {
                p = R_delta * (double)random();
        } while (p > MAX_LOCUS_POS );

        return p;
}

/* Read configuration file, set up data structures. */
init( argc, argv )
int argc;
char *argv[];
{
        int i;
        long int iii;
        double d;
	time_t tloc;

        readconfig();


        /* Set and Test Randomizer */
        iii = 1;
        for ( i = 1; i <= 31; i++ )
                iii = iii * 2;
        iii -= 1;
        d = 1.0 / (double)iii;
        if ( 1 != (int)(d * (double)iii) )
                die( "Bad randomizer" );
        R_delta = d;

        if ( Seed == 0 )
                Seed =  (getpid() * getuid() + (333 * getpid()) ) ^ 999999;

        if ( Log && (NULL == (Logfp = fopen( Logfile, "w+" ) ) ) ) {
                perror( Logfile );
                die( "Can't write log.");
        }
        if ( Log ) {
		time( &tloc );
		fprintf( Logfp, "%s begin %s", Progname, ctime( &tloc ) );
		fprintf( Logfp, "Version %s\n", SCCS );
                fprintf( Logfp, ":Configuration:\n" );
                fprintf( Logfp, "Chromosomes %d\n", N_chromosome );
                fprintf( Logfp, "Loci_per_chromosome %d\n", L_per_c );
                fprintf( Logfp, "Progeny %d\n", N_progeny );
                fprintf( Logfp, "Phase %d\n", Phase_known );
                fprintf( Logfp, "PRNG_seed %d\n", Seed );
		fprintf( Logfp, "Allele_freq %s\n", Allele );
		fprintf( Logfp, "Maternal_chiasma_freq %s\n", M_chiasma );
		fprintf( Logfp, "Paternal_chiasma_freq %s\n", P_chiasma );
        }

	srandom(Seed);
	tree_alloc();
}

/* Allocate TREE structures */
tree_alloc()
{
	int i;
	TREE *new_tree();
	Mom_mom = new_tree();
	Mom = new_tree();
	Dad = new_tree();
	Mom_dad = new_tree();
	Dad_mom = new_tree();
	Dad_dad = new_tree();
	for ( i = 0; i < N_progeny; i++ )
		Progeny[i] = new_tree();
}

TREE *
new_tree()
{
	TREE *t;
	int i;
	if ( NULL == ( t = (TREE *)malloc( sizeof( TREE ) ) ) )
		die( "Out of memory for trees." );
	for (i = 0; i < N_chromosome; i++ ) {
		if ( NULL == ( t->hapmom[i] = malloc( L_per_c + 1 ) ) )
			die( "Out of memory for tree genome." );
		if ( NULL == ( t->hapdad[i] = malloc( L_per_c + 1 ) ) )
			die( "Out of memory for tree genome." );
	}


	return t;

}

readconfig()
{
        FILE *fp;
        char buf[BUFSIZ + BUFSIZ];
        char key[BUFSIZ];
        char val[BUFSIZ];
        int len;
	int i;
        char *s;
        char *p;
	int seed_set;

	if ( 0 != access( Configfile, R_OK ) ) {
		make_config( Configfile );
		fputc( '\007', stderr );
		fprintf( stderr, "\n\nNo configuration file was found!\n" );
		fprintf( stderr,
			"%s has been created for you using default values.\n",
			Configfile );
		fprintf( stderr,
	"You may now edit %s and run %s or just run %s using the defaults.\n\n",
			Configfile, Progname, Progname );
		exit( 0 );
	}
        if ( NULL == ( fp = fopen( Configfile, "r" ) ) ) {
                fprintf( stderr, "No configuration file found.\n" );
                fprintf( stderr, "Working with defaults.\n" );
                return;
        }

	Allele = NULL;
	M_chiasma = NULL;;
	P_chiasma = NULL;;
	L_per_c = -1;  /* Loci per Chromosome */
	Log = -1;
	N_progeny = -1;
	N_chromosome = -1;
	Phase_known = -1;
	Randloci = 0;
	seed_set = 0;
        while( NULL != fgets( buf, (BUFSIZ + BUFSIZ) - 2, fp ) ) {
                len = strlen( buf );
                if ( *buf == '#'  ||  len == 1 ) /* Unix blank lines */
                        continue;
                
                if ( len < 1  ||  len > BUFSIZ ) {
                        usage();
                        die( "Bad Configuration file." );
                }

                if ( buf[len - 1] == '\n' )
                        buf[len - 1] = '\0';

                if ( NULL == ( p = strchr( buf, ':' ) ) ) {
                        usage();
                        die( "Bad Configuration file." );
                }
                *p = '\0';

                strcpy( key, buf );
                len = strlen( p + 1 );
                if ( len < 1  ||  len > BUFSIZ ) {
                        usage();
                        die( "Line too long in configuration file" );
                }
                strcpy( val, p + 1 );
                for( s = key; *s; s++ )
                        *s = tolower( *s );

                if ( 0 == strcmp( key, "progeny" ) )
                        N_progeny = atoi( val );
                else if ( 0 == strcmp( key, "phase" ) )
                        Phase_known = atoi( val );
                else if ( 0 == strcmp( key, "chromosome" ) )
                        N_chromosome = atoi( val );
                else if ( 0 == strcmp( key, "seed" ) ) {
                        Seed = atoi( val );
			seed_set = 1;
		}
                else if ( 0 == strcmp( key, "log" ) )
                        Log = atoi( val );
                else if ( 0 == strcmp( key, "locus" ) )
                        L_per_c = atoi( val );
		else if ( 0 == strcmp( key, "mchiasma" ) ) {
			len = strlen( val );
			for ( i = 0; i < len; i++ )
				if ( ! isdigit(val[i]) )
					die( "chiasma must be digits." );
			if ( NULL == ( M_chiasma = malloc( len + 1 ) ) )
				die( "Can't get memory for mchiasma." );
			strcpy( M_chiasma, val );
		}
		else if ( 0 == strcmp( key, "pchiasma" ) ) {
			len = strlen( val );
			for ( i = 0; i < len; i++ )
				if ( ! isdigit(val[i]) )
					die( "chiasma must be digits." );
			if ( NULL == ( P_chiasma = malloc( len + 1 ) ) )
				die( "Can't get memory for pchiasma." );
			strcpy( P_chiasma, val );
		}
		else if ( 0 == strcmp( key, "allele" ) ) {
			len = strlen( val );
			for ( i = 0; i < len; i++ )
				if ( ! isalnum(val[i]) )
					die( "allele must be alphanumeric." );
			if ( NULL == ( Allele = malloc( len + 1 ) ) )
				die( "Can't get memory for allele." );
			strcpy( Allele, val );
		}
                else if ( 0 == strcmp( key, "randlocus" ) )
                        Randloci = atoi( val );
                else {
                        fprintf( stderr, "%s", key );
                        die( "Illegal configuration file parameter." );
                }
        }
	if ( Allele == NULL ) die ( "Missing allele in configuration file.");
	if ( M_chiasma == NULL) die ("Missing mchiasma in configuration file.");
	if ( P_chiasma == NULL) die ("Missing pchiasma in configuration file.");
	if ( L_per_c < 2  ||  L_per_c > MAX_L_PER_C )
		die ( "Bad locus in configuration file." );
	if ( Log == -1 ) die ("Missing log in configuration file.");
	if ( N_progeny < 1  ||  N_progeny > MAX_PROGENY )
		die ("Bad number of progeny in configuration file." );
	if ( N_chromosome < 1  ||  N_chromosome > MAX_CHROMOSOME )
		die ("Bad number of chromosomes in configuration file." );
	if ( Phase_known == -1 ) die ( "Missing phase in  configuration file.");
	if ( seed_set == 0 ) die ( "Seed not set in configuration file. " );
}

usage()
{
        fprintf( stderr, "Usage: synthdata\n" );

}

/* Print message and exit in error */
die( s ) 
char *s;
{
        fprintf( stderr, "%s\n", s );
        exit( 911 );
}

/*
 *      Return number of cross-overs on this
 *      chromosome at this meiosis.  Chiasma should be all
 *	numbers.
 */
ncross( s )
char *s;
{
        int i, j;
        i = strlen(s);
        j = random() % i;
	i = *(s + j) - '0';
        return i;
}

make_config( fname )
char *fname;
{
	FILE *fp;
	int i;

	static char *text[] = {
	"#",
	"# syd.conf -- Configuration file for syd program.", 
	"#",
	"# Lines beginning with '#' are ignored.",
	"#",
	"# Set number of chromosomes.",
	"chromosome:12",
	"#",
	"# Set number of loci on each chromosome.",
	"# An odd number is better as the two exterme",
	"# loci are always at the telomeres.",
	"locus:11",
	"#",
	"# Set number of progeny in 3rd generation.",
	"progeny:25",
	"#",
	"# Balls in an urn to set maternal crossover frequency.",
	"# At each meiosis, syd will use this table to randomly select",
	"# the number of crossovers.",
	"mchiasma:00000000000000001111111111111111222222223333445",
	"#",
	"# Balls in an urn to set paternal crossover frequency.",
	"# At each meiosis, syd will use this table to randomly select",
	"# the number of crossovers.",
	"pchiasma:00000000000000001111111111111111222222223333445",
	"#",
	"# Balls in an urn to set allele distribution.",
	"# When a grandparent is generated, syd randomly selects",
	"# one of these characters at each allele.",
        "allele:11111111111111112222222222222233333333333344444444555566778",
	"#",
	"# Set phase to 0 for mixed up allele order.  If phase is 1 then",
	"# the maternally contributed allele is always on the left.",
	"phase:1",
	"#",
	"# If seed is 0 program will select a random seed for the PRNG.",
	"# This will produce different data sets each run.",
	"# If seed is non-zero then that seed will be used for the PRNG.",
	"seed:0",
	"#",
	"# If randlocus is 0 then locus positions will be",
	"# established equidistant along the chromosome with ",
	"# one at each telomere.  If randlocus is 1 then",
	"# loci will be at random positions on the chromosomes.",
	"randlocus:0",
	"#",
	"# If log is 0 then syd.log will not be written.",
	"# Set log:100 for maximum information in syd.log.",
	"log:100",
	NULL
	};

        if ( NULL == ( fp = fopen( fname, "w+" ) ) ) {
		perror( fname );
		die( "Couldn't create Configuration file." );
	}

	for( i = 0; text[i] != NULL; i++ )
		fprintf( fp, "%s\n", text[i] );
	
	fclose( fp );
}
